import pathlib
from logging.handlers import QueueHandler
import logging
import sys
import threading
from queue import Queue, Empty
import multiprocessing as mp
    

from piidigger.getmime import getMime
from piidigger import globalfuncs
try:
    from win32api import GetFileAttributes
    win32apiLoaded=True
except ModuleNotFoundError:
    win32apiLoaded=False

from piidigger import classes
from piidigger import console

logger = logging.getLogger(__name__)


def findDirsWorker(config: classes.Config, 
                   queues: dict, 
                   totals: dict,
                   stopEvent: mp.Event,
                  ) -> list:
    '''
    Expects a config object to start from (to get the startDirs paths)

    Populates a queue of all directories that aren't on the excludedDirs list
    '''
    
    
    try:
        ctrlc=False
        logger=logging.getLogger('findDirsWorker')
        logger.addHandler(QueueHandler(queues['logQ']))
        logger.setLevel(config.getLogLevel())
        logger.propagate=False
        
        logger.info('Starting findDirsWorker')

        localQ=list()
        i=0

        for d in config.getStartDirs():
            localQ.append(pathlib.Path(d))
            queues['dirsQ'].put(pathlib.Path(d))
            totals['totalDirs'].value+=1

        while not stopEvent.is_set() or not ctrlc:
            try:
                p=localQ[i]
            except IndexError:
                break
            else:
                i+=1

            
            try:
                for subD in p.iterdir():
                    if stopEvent.is_set():
                        break
                    excludeDir=False
                    try:
                        if subD.is_dir() and not subD.is_symlink():
                            for pattern in config.getExcludeDirs():
                                if str(subD).lower().startswith(pattern.lower()):
                                    logger.debug('Excluding directory %s matched pattern %s', str(subD), pattern)
                                    excludeDir=True
                                    break
                            if not excludeDir:
                                logger.debug('Including directory %s', str(subD))
                                localQ.append(subD)
                                queues['dirsQ'].put(subD)
                                totals['totalDirs'].value+=1
                    except FileNotFoundError:
                        pass
                    except OSError as e:
                        logger.debug('OSError: %s', str(e))
            except PermissionError as e:
                logger.debug('PermissionError: %s', str(e))
            except FileNotFoundError as e:
                logger.debug('FileNotFoundError: %s', str(e))
        
        queues['dirsQ'].put(None)
        globalfuncs.waitOnQ(queues['dirsQ'])
    except KeyboardInterrupt:
        console.normal('\n')
        console.warn('User terminated scan.  Shutting down.')
        logger.info('KeyboardInterrupt received in findDirsWorker')
        globalfuncs.clearQ(queues['dirsQ'])
    finally:
        # All directories have been scanned.  Send the sentinel message to shutdown the consumer threads
        logger.info('Found %d folders', totals['totalDirs'].value)
        logger.info('Stopping findDirsWorker')

        
def findFilesWorker(config: classes.Config, 
                    queues: dict, 
                    totals: dict,
                    stopEvent: mp.Event,
                   ) -> list:
    '''
    Inputs: Config(config), dirsQ as consumer, filesQ as producer

    Places identified files on the filesQ for one of the file handler workers to pick it up
    '''
    
    
    try:
        logger=logging.getLogger('findFilesWorker')
        logger.addHandler(QueueHandler(queues['logQ']))
        logger.setLevel(config.getLogLevel())
        logger.propagate=False

        logger.info('Starting findFilesWorker')

        while True:
            if stopEvent.is_set():
                break
            try:
                item=queues['dirsQ'].get(timeout=.1)
            except Empty:
                continue

            if item==None:
                break
            
            # Path-ify the directory name
            d=pathlib.Path(item)
            
            try: 
                logger.info('Scanning directory: %s', str(d))
                for f in d.iterdir():
                    screenItem=fileChecks(f, config)           
                    if all(screenItem):
                        mimeType = getMime(f)
                        match=fileMatches(f, config.getFileExts(), config.getMimeTypes())
                        if match:
                            queues['filesQ'].put(classes.File(f, mimeType))
                            totals['totalFilesFound'].value+=1
                        else:
                            logger.debug('%s: Item not added (suffix: %s | mime: %s)', f, f.suffix, mimeType)
                    else:
                        logger.debug('%s: Item failed file checks (isFile=%s, isNotZero=%s, isLocalFile=%s)', f, screenItem[0], screenItem[1], screenItem[2])
            except PermissionError as e:
                logger.debug('PermissionError: %s', str(e))
            except FileNotFoundError:
                pass
            except OSError as e:
                logger.debug('OSError: %s', str(e))

        # All files have been identified.  Put the sentinel on the queue for the fileHandlers 
        queues['filesQ'].put(None)
        globalfuncs.waitOnQ(queues['filesQ'])
    except KeyboardInterrupt:
        console.normal('\n')
        console.warn('User terminated scan.  Shutting down.')
        logger.info('KeyboardInterrupt received in findFilesWorker')
        globalfuncs.clearQ(queues['dirsQ'])
        globalfuncs.clearQ(queues['filesQ'])
    finally:
        logger.info('Found %d files', totals['totalFilesFound'].value)
        logger.info('Stopping findFilesWorker')


def fileChecks(f: pathlib.Path, 
               config: classes.Config) -> tuple:
    '''
    Performs initial checks to see if the item is a file, has content, and is local (e.g. OneDrive file already local file system)
    Returns a tuple of (isFile, isNotZero, isLocalFile)
    '''
    recallMask=0x400000
    offlineMask=0x1000

    isFile = f.is_file() 
    isNotZero = f.stat().st_size > 0

    # Checks if the file is local.  Non-local OneDrive and Dropbox files will be skipped.
    # https://superuser.com/questions/1718444/determining-if-a-onedrive-file-is-synced-locally-via-a-terminal
    # For OneDrive, the 0x400000 (Recall) bit will be set (TRUE) if the file isn't on the disk.
    # For Dropbox, the 0x1000 (Offline) bit will be set (TRUE) if the file isn't on the disk.
    # According to the linked article, this might not be derived from MS documentation, so it could break in the future, but from my own testing, it seems to indicate correctly
    # And the Dropbox behavior is from direct observation
    if win32apiLoaded and config.getLocalFilesOnly():
        try:
            attr=GetFileAttributes(str(f))
        except Exception:
            # The most common exception is that the file is in use, which means its local. But, when in doubt, assume it's a local file.
            isLocalFile=True
        else:
            oneDriveRemote=bool(attr & recallMask)
            dropboxRemote=bool(attr & offlineMask)

            # A file is only remote when either the RECALL or the Offline bit is set.  We need the inverse of that.
            isLocalFile=not(oneDriveRemote or dropboxRemote)
    else:
        isLocalFile=True

    return (isFile, isNotZero, isLocalFile)

def fileMatches(f: pathlib.Path, fileExts: list, mimeTypes: list) -> bool:
    extFound = f.suffix in fileExts
    mimeFound = getMime(f) in mimeTypes

    return extFound or mimeFound


def _filesQStubWorker(queues: dict, totals: dict):
    import random, time
    
    result={'filename': '/path/to/filename.txt',
            'pan': {
                'visa': {'4893 01** **** 6137',}}}
    
    logger=logging.getLogger('filesQStubWorker')
    logger.addHandler(QueueHandler(queues['logQ']))
    logger.propagate=False
    logger.info('Starting filesQStubWorker')

    while True:
        item=queues['filesQ'].get()
        if item == None:
            queues['filesQ'].put(item)
            break
        r=(random.random())
        time.sleep(r/4)
        totals['totalFilesScanned'].value+=1
        if r<.1:
            queues['resultsQ'].put(result)
    
    # Send the sentinel message to the results queue to shutdown
    queues['resultsQ'].put(None)
    logger.info('Stopping filesQStubWorker')

def _resultsQStubWorker(queues: dict, totals: dict):
    logger=logging.getLogger('resultsQStubWorker')
    logger.addHandler(QueueHandler(queues['logQ']))
    logger.propagate=False
    logger.info('Starting resultsQStubWorker')

    while True:
        item=queues['resultsQ'].get()
        if item == None:
            break
        totals['totalResults'].value+=1

    logger.info('Stopping filesQStubWorker')

def main():
    '''
    Testing for the fileparser module.  Normally these options would all be set in the main program or passed to findDirsWorker and findFilesWorker
    '''
    from datetime import datetime
    import os
    start=datetime.now()
    config=classes.Config('dne.toml')
    queues={name: mp.Queue() for name in ['logQ', 'dirsQ', 'filesQ', 'resultsQ']}
    totals={k: mp.Value('i', 0) for k in ['totalDirs', 'totalFilesFound', 'totalFilesScanned', 'totalResults']}
    stopEvent=mp.Event()
    stopEvent.clear()
    
    
    console.normal('File spec to match: %s' % (str(config.getFileExts())))
    console.normal('Mime types to match: %s' % (str(config.getMimeTypes())))
    console.normal('Scanning %s for matching files...' % (config.getStartDirs()))

    #####################################
    # Bring up the pipe line in reverse order -- starting with results and working back to scanning for directories
    #####################################
    processResultsThread=threading.Thread(target=_resultsQStubWorker, args=(queues,totals,), daemon=True)
    processResultsThread.start()

    filesQProcesses=[mp.Process(target=_filesQStubWorker, args=(queues,totals)) for i in range(os.cpu_count())]
    
    #Uncomment for debugging the files queue stub
    #filesQProcesses=[mp.Process(target=_filesQStubWorker, args=(queues,totals))]

    # for process in filesQProcesses:
    #     try:
    #         process.start()
    #     except KeyboardInterrupt:
    #         print('Quitting...')
    #         exit()

    # findFilesThread=threading.Thread(target=findFilesWorker, args=(config, queues, totals))
    # findFilesThread.start()
    
    findDirsThread=threading.Thread(target=findDirsWorker, args=(config, queues, totals, stopEvent))
    findDirsThread.start()

    # Populate the dirsQ with the initial directories from the config.  findFilesWorker will start with these while waiting for findDirs to identify more
    # for item in config.getStartDirs():
    #     queues['dirsQ'].put(item)

    ##################################
    # Wait for the threads/processes to complete by joining each of the threads/processes in the pipeline in pipeline order
    ##################################
    findDirsThread.join()
    # findFilesThread.join()
    # for process in filesQProcesses:
    #     process.join()
    # processResultsThread.join()

    # There will be an extra item in the files queue, so pull it out.  Without this, the queuefeederthread used by MP will not shutdown
    _=queues['filesQ'].get()

    print("\nTotal time: ", datetime.now() - start)

    sys.exit()
    
if __name__ == '__main__':
    main()



from multiprocessing import Queue
from queue import Empty
import os
import ctypes
import platform
from time import sleep


from piidigger import console
from piidigger import filehandlers as fh
from piidigger import datahandlers as dh
from piidigger import outputhandlers as oh

######################################################
############### Global variables #####################
######################################################

errorCodes={
    "ok": 0,
    "userTerm": 1,
    "invalidConfig": 1,
    'unknownError': 1024,
}

excelBlankRowLimit=250
excelBlankColLimit=500

# Dynamically build the supported file handlers based on the contents of the filehandlers package.
# Each file handler needs a globally-defined variable called "handles" with a dictionary as follows:
#   ext: [a list of file extensions with the leading .]
#   mime: [a list of mime-type strings]

fileHandlers={ handler: getattr(getattr(fh, handler), 'handles') for handler in fh.__dir__() if not handler.startswith('__') }

# Leave these in here as reference until I write more file handlers.
# fileHandlers={
#     'archive': {
#         'ext': ['.7z','.bz2','.gz','.gzip','.tar','.zip',],
#         'mime': ['application/x-7z-compressed','application/gzip','application/zip','application/x-tar','application/x-bzip2',],
#     },
#     'appledocs': {
#         'ext': ['.numbers',],
#         'mime': [],
#     },
#     'csv': {
#         'ext': ['.csv',],
#         'mime': ['text/csv'],
#     },
#     'googledocs': {
#         'ext': ['.gsheet',],
#         'mime': [],
#     },
#     'msoffice': {
#         'ext': ['.doc','.docx','.ppt','.pptx','.xls','.xlsx',],
#         'mime': [
#             'application/msword',
#             'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
#             'application/vnd.ms-powerpoint',
#             'application/vnd.openxmlformats-officedocument.presentationml.presentation',
#             'application/vnd.ms-excel',
#             'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
#         ],
#     },
#     'openoffice': {
#         'ext': ['.odp','.ods','.odt','.odp',],
#         'mime': ['application/vnd.oasis.opendocument.spreadsheet', 'application/vnd.oasis.opendocument.text', 'application/vnd.oasis.opendocument.presentation'],
#     },
#     'outlook': {
#         'ext': ['.pst','.ost',],
#         'mime': [],
#     },
#     'pdf': {
#         'ext': ['.pdf',],
#         'mime': ['application/pdf',],
#     },
#     'richtext': {
#         'ext': ['.rtf',],
#     	'mime': ['text/rtf'],
#     },    
# }


version='0.0.1'

######################################################
############### Global functions #####################
######################################################


def clearQ(q: Queue):
    '''Clears a queue of all contents'''
    while True:
        try:
            _=q.get(block=False)
        except Empty:
            break
        except KeyboardInterrupt:
            continue 
    

def waitOnQ(q: Queue):
    while not q.empty():
        sleep(.01)
    

def countResults(results: dict) -> int:
    '''Receives a dictionary of results and returns the total match count across all match sets in the dictionary'''

    count=0
    for key in results.keys():
        item=results[key]
        if isinstance(item, dict):
            count+=countResults(item)
        elif isinstance(item, list):
            count+=len(item)
    
    return count


def getAllDataHandlerModules() -> list:
    '''
    Returns a list containing the data handler modules from datahandler package
    '''
    
    return [getattr(dh, module) for module in getSupportedDataHandlerNames()]


def getDataHandlerModule(name: str):
    '''
    Returns a module object for a specific data handler
    '''
    try:
        return getattr(dh, name)
    except Exception:
        return None


def getDefaultConfig() -> dict:
    '''
    Returns a default configuration object for when a configuration file couldn't be found.
    '''

    return {
        'dataHandlers': [
            'pan',
        ],
        'localFilesOnly': True,
        'results': {'path': "piidigger-results/",
                    'json': True,
                    'text': True,}, 
        'includeFiles': {'ext': 'all', 
                            'mime': 'all', 
                            'startDirs': {'windows': 'all', 
                                        'linux': ['/'], 
                                        'darwin': ['/']}},
        'excludeDirs': {
            'windows': ['C:\\Windows', 'C:\\Program Files (x86)', 'C:\\Program Files',], 
            'linux': ['/boot', '/dev', '/etc', '/proc', '/run', '/snap', '/sys', '/usr/bin', '/usr/lib', '/usr/lib32', '/usr/lib64', '/usr/libx32', '/usr/local', '/usr/sbin', '/usr/share', '/usr/src/', '*/.vscode-server', '/mnt/c', '/mnt/d', '/mnt/wslg', '/wsl'],
            'darwin': ["/dev", '/etc', '/usr/bin', '/usr/local/Homebrew', '/usr/lib', '/usr/sbin', '/Applications', '/Library/Developer', '/Library/Documentation', '/System',]},
        'logging': {'logLevel': 'INFO', 
                    'logFile': 'logs/piidigger.log'}
        }


def getEnabledDataHandlerModules(moduleNames: list):
    '''
    Receives a list of enabled module names and returns a list of the module objects
    '''
    return [getDataHandlerModule(name) for name in moduleNames]


def getFileHandlerName(ext: str, mime: str) -> str:
    '''
    Receives a file extension and MIME type

    Returns a file handler to use to work with the file or None

    All file handlers should be in the "handler" module/directory in a Python file by the name of handler
    '''

    handler=None

    for key in fileHandlers.keys():
        if (ext in fileHandlers[key]['ext']) or (mime in fileHandlers[key]['mime']):
            handler=str(key)
            break
    
    return handler


def getFileHandlerModule(name):
    '''
    Returns a module object for a file handler by the provided name
    '''    
    try:
        return getattr(fh, name)
    except Exception:
        return None


def getOutputHandlerModule(name: str):
    '''
    Returns a module object for a specific output handler
    '''

    try:
        return getattr(oh, name)
    except Exception:
        return None    


def getOSType() -> str:
    return platform.system().lower()


def getSupportedDataHandlerNames() -> list:
    return [n for n in dh.__dir__() if not n.startswith("__")]


def getSupportedFileExts() -> list:
    '''
    Returns a list of all supported file extentions
    '''

    exts=[]

    for key in fileHandlers.keys():
        exts+=fileHandlers[key]['ext']

    return exts


def getSupportedFileMimes() -> list:
    '''
    Returns a list of all supported file extentions
    '''

    mimes=[]

    for key in fileHandlers.keys():
        mimes+=fileHandlers[key]['mime']

    return mimes


def isAdmin() -> bool:
    try:
        check = (os.geteuid() == 0)
    except AttributeError:
        check = ctypes.windll.shell32.IsUserAnAdmin() != 0
    return check


def makeChunks(s: str, chunkSize: int=650) -> list:
    '''Breaks up a string into smaller strings not larger than chunkSize'''

    words = s.split()
    wordNum = 0
    chunks=list()

    while wordNum<len(words):
        chunk=str()
        while len(chunk) < chunkSize and wordNum < len(words):
            if len(words[wordNum]) > chunkSize:
                # Break up super-long strings into shorter words and add them to the word list
                chunkList = [words[wordNum][i:i+chunkSize] for i in range(0, len(words[wordNum]), chunkSize)]
                words = [*words, *chunkList]
            else:
                # Add the next word to the current chunk
                chunk+=words[wordNum] + ' '
            wordNum+=1
            
        # Add the current chunk to the list of chunks
        chunks+=[chunk.strip()]

    return chunks


def processMatches(results: dict, 
                   matches: dict, 
                   dhName: str) -> dict:
    
    '''Process the results from RE matches and add them to the results dictionary'''

    for key in matches.keys():
        value = matches[key]
        if value:
            if not dhName in results['matches']:
                results['matches'][dhName]=dict()
            if not key in results['matches'][dhName]:
                results['matches'][dhName][key]=set()
            results['matches'][dhName][key].update(value)

    return results


def progressLine(*pargs, **kwargs):
    '''
    Prints a status line that includes details about directories, files, and results
    '''
    screenWidth=console.getTerminalSize()[0]

    line='Folders scanned: %d | Files identified: %d | Files scanned: %d | Results found: %d' % (kwargs['totalDirs'].value, kwargs['totalFilesFound'].value, kwargs['totalFilesScanned'].value, kwargs['totalResults'].value)

    if len(line) > screenWidth:
        line=line[:screenWidth-1]

    console.status(line)


def sizeof_fmt(num, suffix="B"):
    '''
    Returns a human-readable string for bytes.
    '''

    # Taken from https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size
    for unit in ("", "K", "M", "G", "T", "P", "E", "Z"):
        if abs(num) < 1024.0:
            return f"{num:.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Y{suffix}"


def writeDefaultConfig(tomlFile: str):
    # This is a total kluge, but without a reasonable Python library to write a TOML v1.1 file based on a Python dictionary, we have to build the default config file from scratch
    
    def _tomlfy(key, value):
        if isinstance(value, str):
            return key + ' = "' + value + '"'
        elif isinstance(value, bool):
            return key + ' = ' + str(value).lower()
        else:
            return key + ' = ' + str(value).replace('\'', '"')
    
    defaultConfig=getDefaultConfig()
    lines = list()
    for key in ['dataHandlers',]:
        lines.append(_tomlfy(key, defaultConfig[key]))

    lines.append('')
    for key in ['localFilesOnly']:
        lines.append(_tomlfy(key, defaultConfig[key]))
    
    lines.append('')
    lines.append('[results]')
    for key in ['path', 'json', 'text']:
        lines.append(_tomlfy(key, defaultConfig['results'][key]))
        
    lines.append('')
    lines.append('[includeFiles]')
    for key in ['ext', 'mime']:
        lines.append(_tomlfy(key,defaultConfig['includeFiles'][key]))
            
    lines.append('')
    lines.append('[includeFiles.startDirs]')
    for key in ['windows', 'linux', 'darwin']:
        lines.append(_tomlfy(key, defaultConfig['includeFiles']['startDirs'][key]))
        # if isinstance(value, str):
        #     lines.append(pf + ' = "' + value + '"')
        # else:
        #     lines.append(pf + ' = ' + str(defaultConfig['includeFiles']['startDirs'][pf]).replace('\'', '"'))
    
    lines.append('')
    lines.append('[excludeDirs]')
    for key in ['windows', 'linux', 'darwin']:
        lines.append(_tomlfy(key, defaultConfig['excludeDirs'][key]))
    
    lines.append('')
    lines.append('[logging]')
    for key in ['logLevel', 'logFile']:
        lines.append(_tomlfy(key, defaultConfig['logging'][key]))
    
    try:
        with open(tomlFile, 'w') as tf:
            tf.writelines(line + '\n' for line in lines)
        return "Success"
    except Exception as e:
        return str(e)




import pytest

from src.piidigger.getencoding import getEncoding

@pytest.mark.parametrize('testFile, expected_result', [
                                ('testdata/pan/sample-pans.json', 'ascii'),
                                ('testdata/binary-json.json', None),
                            ]
                        )
def test_getEncoding(testFile, expected_result):
    result=getEncoding(testFile)

    assert result==expected_result